
"""
$(TYPEDEF)

Struct to hold sparse matrix in the linked list format.

Modeled after the linked list sparse matrix format described in 
the  [whitepaper](https://www-users.cs.umn.edu/~saad/software/SPARSKIT/paper.ps)
and the  [SPARSEKIT2 source code](https://www-users.cs.umn.edu/~saad/software/SPARSKIT/SPARSKIT2.tar.gz)
by Y. Saad. He writes "This is one of the oldest data structures used for sparse matrix computations."

The relevant source [formats.f](https://salsa.debian.org/science-team/sparskit/blob/master/FORMATS/formats.f)
is also available in the debian/science gitlab.

Probably this format was around already in SPARSPAK by E.Chu, A.George and J.Liu, however this is 
hard to verify, as it indeed appears that the source code of SPARSPAK [vanished from the internet](http://www.netlib.org/sparspak/readme).

The advantage of the linked list structure is the fact that upon insertion
of a new entry, the arrays describing the structure grow at their respective ends and
can be conveniently updated via `push!`.  No copying of existing data is necessary.

$(TYPEDFIELDS)
"""
mutable struct SparseMatrixLNK{Tv,Ti<:Integer} <: AbstractSparseMatrix{Tv,Ti}

    """
    Number of rows
    """
    m::Ti


    """
    Number of columns
    """
    n::Ti


    """
    Number of nonzeros
    """
    nnz::Ti

    """
    Length of arrays
    """
    nentries::Ti


    """
    Linked list of column entries. Initial length is n,
    it grows with each new entry.
    
    colptr[index] contains the next
    index in the list or zero, in the later case terminating the list which
    starts at index 1<=j<=n for each column j.
    """
    colptr::Vector{Ti}      


    """
    Row numbers. For each index it contains the zero (initial state)
    or the row numbers corresponding to the column entry list in colptr.

    Initial length is n,
    it grows with each new entry.
    """
    rowval::Vector{Ti}

    """
    Nonzero entry values correspondin to each pair
    (colptr[index],rowval[index])

    Initial length is n,  it grows with each new entry.
    """
    nzval::Vector{Tv}       
end



"""
$(SIGNATURES)
    
Constructor of empty matrix.
"""
SparseMatrixLNK{Tv,Ti}(m::Integer, n::Integer)  where {Tv,Ti<:Integer} =    SparseMatrixLNK{Tv,Ti}(m,n,0,n,zeros(Ti,n),zeros(Ti,n),zeros(Tv,n))

"""
$(SIGNATURES)
    
Constructor of empty matrix.
"""
SparseMatrixLNK(valuetype::Type{Tv},indextype::Type{Ti},m::Integer, n::Integer)  where {Tv,Ti<:Integer} =SparseMatrixLNK{Tv,Ti}(m,n)

"""
$(SIGNATURES)
    
Constructor of empty matrix.
"""
SparseMatrixLNK(valuetype::Type{Tv},m::Integer, n::Integer)  where {Tv} =SparseMatrixLNK(Tv,Int,m,n)

"""
$(SIGNATURES)
    
Constructor of empty matrix.
"""
SparseMatrixLNK(m::Integer, n::Integer)  where {Tv} =SparseMatrixLNK(Float64,m,n)


"""
$(SIGNATURES)
    
Constructor from SparseMatrixCSC.

"""
function SparseMatrixLNK(csc::SparseArrays.SparseMatrixCSC{Tv,Ti})  where {Tv,Ti<:Integer}
    lnk=SparseMatrixLNK{Tv,Ti}(csc.m,csc.n)
    for j=1:csc.n
        for k=csc.colptr[j]:csc.colptr[j+1]-1
            lnk[csc.rowval[k],j]=csc.nzval[k]
        end
    end
    lnk
end





"""
$(SIGNATURES)
    
Return value stored for entry or zero if not found
"""
function Base.getindex(lnk::SparseMatrixLNK{Tv,Ti},i::Integer, j::Integer) where {Tv,Ti<:Integer}

    if !((1 <= i <= lnk.m) & (1 <= j <= lnk.n))
        throw(BoundsError(lnk, (i,j)))
    end
    
    k=j
    while k>0
        if lnk.rowval[k]==i
            return lnk.nzval[k]
        end
        k=lnk.colptr[k]
    end

    return zero(Tv)
end




"""
$(SIGNATURES)
    
Update value of existing entry, otherwise extend matrix.
"""
function Base.setindex!(lnk::SparseMatrixLNK{Tv,Ti}, _v, _i::Integer, _j::Integer) where {Tv,Ti<:Integer}
    v = convert(Tv, _v)
    i = convert(Ti, _i)
    j = convert(Ti, _j)

    if !((1 <= i <= lnk.m) & (1 <= j <= lnk.n))
        throw(BoundsError(lnk, (i,j)))
    end
    
    # Set the first  column entry if it was not yet set.
    if lnk.rowval[j]==0
        lnk.rowval[j]=i       
        lnk.nzval[j]=v
        lnk.nnz+=1
        return lnk
    end

    # Traverse list for existing entry
    k=j
    k0=j
    while k>0
        # Update value and return if entry has been found
        if lnk.rowval[k]==i
            lnk.nzval[k]=v
            return lnk
        end
        k0=k
        # Next element in the list
        k=lnk.colptr[k]
    end

    # increase number of entries
    lnk.nentries+=1
    if length(lnk.nzval)<lnk.nentries
        newsize=Int64(ceil(5.0*lnk.nentries/4.0))
        resize!(lnk.nzval,newsize)
        resize!(lnk.rowval,newsize)
        resize!(lnk.colptr,newsize)
    end
    
    # Append entry if not found
    lnk.nzval[lnk.nentries]=v
    lnk.rowval[lnk.nentries]=i

    # Shift the end of the list
    lnk.colptr[lnk.nentries]=0
    lnk.colptr[k0]=lnk.nentries

    # Update number of nonzero entries
    lnk.nnz+=1
    return lnk
end


"""
$(SIGNATURES)

Return tuple containing size of the matrix.
"""
Base.size(lnk::SparseMatrixLNK) = (lnk.m, lnk.n)


"""
$(SIGNATURES)

Return number of nonzero entries.
"""
SparseArrays.nnz(lnk::SparseMatrixLNK)=lnk.nnz


"""
$(SIGNATURES)

Dummy flush! method for SparseMatrixLNK. Just
used in test methods
"""
function flush!(lnk::SparseMatrixLNK{Tv, Ti}) where{Tv, Ti}
    return lnk
end






# Struct holding pair of value and row
# number, for sorting
struct ColEntry{Tv,Ti<:Integer}
    rowval::Ti
    nzval::Tv
end

# Comparison method for sorting
Base.isless(x::ColEntry{Tv, Ti},y::ColEntry{Tv, Ti}) where {Tv,Ti<:Integer} = (x.rowval<y.rowval)

"""
$(SIGNATURES)

Add SparseMatrixCSC matrix and [`SparseMatrixLNK`](@ref)  lnk.
"""
function Base.:+(lnk::SparseMatrixLNK{Tv,Ti},csc::SparseMatrixCSC{Tv,Ti}) where {Tv,Ti<:Integer}
    @assert(csc.m==lnk.m)
    @assert(csc.n==lnk.n)

    xnnz=nnz(csc)+nnz(lnk)
    colptr=Vector{Ti}(undef,csc.n+1)
    rowval=Vector{Ti}(undef,xnnz)
    nzval=Vector{Tv}(undef,xnnz)

    # Detect the maximum column length of lnk
    lnk_maxcol=0
    for j=1:csc.n
        lcol=zero(Ti)
        k=j
        while k>0
            lcol+=1
            k=lnk.colptr[k]
        end
        lnk_maxcol=max(lcol,lnk_maxcol)
    end
    
    # pre-allocate column  data
    col=[ColEntry{Tv,Ti}(0,0) for i=1:lnk_maxcol]


    
    inz=1 # counts the nonzero entries in the new matrix

    # loop over all columns
    for j=1:csc.n

        # Copy extension entries into col and sort them
        k=j
        l_lnk_col=zero(Ti)
        while k>0
            if lnk.rowval[k]>0
                l_lnk_col+=1
                col[l_lnk_col]=ColEntry(lnk.rowval[k],lnk.nzval[k])
            end
            k=lnk.colptr[k]
        end
        sort!(col,1,l_lnk_col, Base.QuickSort, Base.Forward)


        # jointly sort lnk and csc entries  into new matrix data
        colptr[j]=inz
        jlnk=one(Ti) # counts the entries in col
        jcsc=csc.colptr[j]  # counts entries in csc

        in_csc_col(jcsc)=(nnz(csc)>zero(Ti)) && (jcsc<csc.colptr[j+1])
        in_lnk_col(jlnk)=(jlnk<=l_lnk_col)

        while true
            if in_csc_col(jcsc) &&  (in_lnk_col(jlnk)  && csc.rowval[jcsc]<col[jlnk].rowval  ||  !in_lnk_col(jlnk))
                # Insert entries from csc into new structure
                rowval[inz]=csc.rowval[jcsc]
                nzval[inz]=csc.nzval[jcsc]
                jcsc+=1
                inz+=1
            elseif in_csc_col(jcsc) &&  (in_lnk_col(jlnk)  && csc.rowval[jcsc]==col[jlnk].rowval)
                # Add up entries from csc and lnk
                rowval[inz]=csc.rowval[jcsc]
                nzval[inz]=csc.nzval[jcsc]+col[jlnk].nzval
                jcsc+=1
                inz+=1
                jlnk+=1
            elseif in_lnk_col(jlnk)
                # Insert entries from lnk res. col into new structure
                rowval[inz]=col[jlnk].rowval
                nzval[inz]=col[jlnk].nzval
                jlnk+=1
                inz+=1
            else 
                break
            end
        end
    end
    colptr[csc.n+1]=inz
    SparseMatrixCSC{Tv,Ti}(csc.m,csc.n,colptr,rowval,nzval)
end

Base.:+(csc::SparseMatrixCSC{Tv,Ti},lnk::SparseMatrixLNK{Tv,Ti}) where {Tv,Ti<:Integer}=lnk+csc

"""
$(SIGNATURES)
    
Constructor from SparseMatrixCSC.

"""
function SparseArrays.SparseMatrixCSC{Tv,Ti}(lnk::SparseMatrixLNK{Tv,Ti})  where {Tv,Ti<:Integer}
    csc=spzeros(lnk.m,lnk.n)
    lnk+csc
end
